# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright(C) 2022 Jakob Knollmüller, Philipp Arras

import nifty8 as ift
import numpy as np
import resolve as rve
from matplotlib import pyplot as plt
import sys

import util as u

try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    master = comm.Get_rank() == 0
except ImportError:
    comm = None
    master = True

ms, iobs = sys.argv[1:]
obs = rve.ms2observations_all(ms, "DATA")
obs = list(obs)[int(iobs)]
# obs.save("obs", compress=True)
# obs = rve.Observation.load("obs")

print(f"Reconstructing {obs.source_name}")

space_fov = rve.str2rad("1.5amin")
space_npix = 256
dx = space_fov / space_npix
sdom = ift.RGSpace([space_npix, space_npix], [dx, dx])

correlated_field = ift.SimpleCorrelatedField(
    sdom,
    offset_mean=22,
    offset_std=(1e-3, 1e-6),
    fluctuations=(1.0, 0.8),
    loglogavgslope=(-4.0, 0.5),
    flexibility=None,
    asperity=None,
)
# See webpage for meaning of prior parameters
# https://ift.pages.mpcdf.de/nifty/nifty8/user/getting_started_4_CorrelatedFields.html

position_space = rve.default_sky_domain(sdom=sdom)
sky = correlated_field.exp().ducktape_left(position_space)

# R = rve.InterferometryResponse(obs, position_space, do_wgridding=False, epsilon=1e-6)
# print(R(sky(ift.from_random(sky.domain))).val)
# print(obs.vis)

likelihood = rve.ImagingLikelihood(obs, sky, epsilon=1e-6, do_wgridding=False, nthreads=1)

if master:
    dirty = rve.dirty_image(obs, "uniform", position_space, do_wgridding=False, epsilon=1e-6, nthreads=4)
    plt.imsave("dirty_uniform.png", dirty.val.squeeze().T, origin="lower")

# Minimization parameters
ic_sampling = ift.AbsDeltaEnergyController(
    name="Sampling (linear)", deltaE=0.05, iteration_limit=100
)
ic_newton = ift.AbsDeltaEnergyController(
    name="Newton", deltaE=0.5, convergence_level=2, iteration_limit=10
)
ic_lbfgs = ift.AbsDeltaEnergyController(
    name="lbfgs", deltaE=0.5, convergence_level=2, iteration_limit=20
)
minimizer_newton = ift.NewtonCG(ic_newton)
minimizer_lbfgs = ift.VL_BFGS(ic_lbfgs)

# Minimize KL
n_iterations = 20
n_samples = 3
minimizer = minimizer_newton


def inspect_callback(sl, iglobal):
    sky_mean = sl.average(sky).val.T
    if master:
        plt.imsave(
            f"sky_{iglobal}.png", sky_mean, origin="lower", cmap="afmhot"
        )


samples = ift.optimize_kl(
    likelihood,
    n_iterations,
    n_samples,
    minimizer,
    ic_sampling,
    None,
    export_operator_outputs={"sky": sky.ducktape_left(sdom)},
    output_directory="results",
    inspect_callback=inspect_callback,
    comm=comm,
)
